
"""
    Genetic algorithm for the point scattering problem
    (maximize the minimum distance for points in the unit circle)
    
    
    David Mathias
    November 2021
    
"""

import random
import math
from copy import deepcopy
from time import time
import matplotlib.pyplot as plt

from point import Point
from individual import Individual
from population import Population

from constants import NUM_POINTS, R_FACTOR, T_FACTOR, CXPB, MUTPB, MUT_INDPB, GENS, POP_SIZE, ELITE_NUM


# main GA function
def evolve(population, cxpb, mutpb, ngen):
         
    # evaluate all individuals
    population.evaluate()
    
    # sort population and print fitness for best individual
    population.sort(key=lambda x: x.fitness, reverse=True)
    print('generation: {:6d}   best individual: {}'.format(0, population.pop[0].fitness))
    
    
    # Begin the generational process
    for gen in range(1, ngen + 1):
        
        population.shuffle()
        
        # Select individuals to serve as parents by copying current parent population
        offspring = deepcopy(population)
        
        # crossover
        # if the parents don't mate, mutate them
        for child1, child2 in zip(offspring.pop[0::2], offspring.pop[1::2]):
            if random.random() < CXPB:
                child1.crossover(child2)
                child1.fitness = -1
                child2.fitness = -1
            else:
                child1.mutate(MUT_INDPB)
                child2.mutate(MUT_INDPB)
                child1.fitness = -1
                child2.fitness = -1
                
                        
        # Evaluate the newly created individuals
        offspring.evaluate()

        # replacement
        # currently chooses ELITE_NUM best individuals and fills rest of
        # next_pop with individuals chosen at random
        population.pop.extend(offspring.pop)
        del offspring
        
        if ELITE_NUM > 0:
            population.sort(key=lambda x: x.fitness, reverse=True)
            next_pop = population.pop[:ELITE_NUM]
            del population.pop[:ELITE_NUM]
        else:
            next_pop = []
        
        # randomize the population so that we choose a random sample
        # for the next generation
        population.shuffle()
        next_pop.extend(population.pop[:POP_SIZE - len(next_pop)])
            
        # deepcopy the population
        population = Population(next_pop)
        del next_pop
                                
        # periodically output progress report
        if gen % 100 == 0:
            population.sort(key=lambda x: x.fitness, reverse=True)
            print('generation: {:6d}   best individual: {}'.format(gen, population.pop[0].fitness))
            
    return population


    
def main():
    random.seed(time())

    pop = Population()
    
    print('\nUnit Circle:')
    print('   Number of points: {}'.format(NUM_POINTS))
    print('   Population size: {}'.format(POP_SIZE))
    print('   Generations: {}\n'.format(GENS))
    
    pop  = evolve(pop, cxpb=CXPB, mutpb=MUTPB, ngen=GENS)
        
    # sort population
    pop.sort(key=lambda x: x.fitness, reverse=True)
        
    indiv = pop.pop[0]
    
    # output the best individual
    print()
    print(indiv)
    
    # for best individual, get the min distance and the indices of
    # the points for that distance
    dist, pt1, pt2 = indiv.min_dist()
    print()
    print('Best min distance: {}  pt1: {}  pt2: {}'.format(dist, pt1, pt2))
    print()
    
    # for plotting the results
    # create list of angles and list of radii
    theta = [indiv.points[i].theta for i in range(NUM_POINTS)]
    r = [indiv.points[i].radius for i in range(NUM_POINTS)]
    
    # plot the points and draw a line between closest pair
    ax = plt.subplot(111, projection='polar', zorder=10)
    c = ax.scatter(theta, r)
    ax.plot((theta[pt1], theta[pt2]), (r[pt1], r[pt2]), c='#ff0000', zorder=0)
    for i in range(NUM_POINTS):
        ax.annotate(i, xy=(theta[i], r[i]))
    ax.set_rmax(1.1)
    
    plt.show()
    

if __name__ == "__main__":
    main()

